package de.hhhammer.playtime.ng.persistence;

import de.hhhammer.playtime.ng.player.Player;
import de.hhhammer.playtime.ng.serialization.SerializationFactory;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.nio.file.Files;
import java.util.ArrayList;
import java.util.List;
import java.util.Optional;
import java.util.UUID;

public final class FileSystemPlayerTimeDB implements PlayerTimeDB {

    private final File saveDirectory;
    private final SerializationFactory serializationFactory;

    public FileSystemPlayerTimeDB(final File saveDirectory, final SerializationFactory serializationFactory) {
        this.saveDirectory = saveDirectory;
        this.serializationFactory = serializationFactory;
    }

    @Override
    public void flush(final Player player) throws WritePlayerException {
        final var playerFile = new File(saveDirectory, player.uuid().toString());
        final String serializedPlayer = serializationFactory.createSerializer(player).serialize();
        try (final var writer = new FileWriter(playerFile)) {
            writer.write(serializedPlayer);
            // ensure everything is written
            writer.flush();
        } catch (IOException e) {
            throw new WritePlayerException("Could not persist player data from player: " + player.uuid(), e);
        }
    }

    @Override
    public Optional<Player> findById(final UUID uuid) throws ReadPlayerException {
        final File[] files = getFiles();
        for (final var file : files) {
            if (!file.getName().equals(uuid.toString())) {
                continue;
            }

            final Player player = readPlayerFromFile(file);
            return Optional.of(player);
        }

        return Optional.empty();
    }

    @Override
    public Optional<Player> findByName(final String name) throws ReadPlayerException {
        final File[] files = getFiles();
        for (final var file : files) {
            final Player player = readPlayerFromFile(file);
            if (player.name().equals(name)) {
                return Optional.of(player);
            }
        }

        return Optional.empty();
    }

    @Override
    public List<Player> findAll() throws ReadPlayerException {
        final File[] files = getFiles();
        final var playerList = new ArrayList<Player>();
        for (final var file : files) {
            final Player player = readPlayerFromFile(file);
            playerList.add(player);
        }

        return playerList;
    }

    private File[] getFiles() {
        final File[] files = saveDirectory.listFiles();
        if (files == null) {
            throw new RuntimeException("Could not find files in: " + saveDirectory.getAbsolutePath());
        }

        return files;
    }

    private Player readPlayerFromFile(final File playerFile) throws ReadPlayerException {
        try {
            final String fileContent = Files.readString(playerFile.toPath());
            return serializationFactory.createDeserializer(fileContent);
        } catch (IOException e) {
            final String fileName = playerFile.getName();
            throw new ReadPlayerException("Could not read file: " + fileName, e);
        }
    }

}
