package de.hhhammer.playtime.ng.persistence;

import de.hhhammer.playtime.ng.player.Player;
import de.hhhammer.playtime.ng.serialization.SerializationFactory;
import de.hhhammer.playtime.ng.serialization.serializer.Serializer;
import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.extension.ExtendWith;
import org.mockito.Mock;
import org.mockito.junit.jupiter.MockitoExtension;

import java.io.File;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.List;
import java.util.Optional;
import java.util.UUID;

import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.ArgumentMatchers.anyString;
import static org.mockito.Mockito.*;

@ExtendWith(MockitoExtension.class)
class FileSystemPlayerTimeDBTest {

    @Mock
    private SerializationFactory mockSerializationFactory;
    @Mock
    private Player mockPlayer;
    @Mock
    private Serializer mockSerializer;
    private File testDirectory;
    private FileSystemPlayerTimeDB fileSystemDB;

    @BeforeEach
    void setup() throws IOException {
        final Path path = Files.createTempDirectory("playtime").toAbsolutePath();
        this.testDirectory = path.toFile();
        this.fileSystemDB = new FileSystemPlayerTimeDB(testDirectory, mockSerializationFactory);
    }

    @AfterEach
    void cleanup() {
        testDirectory.delete();
    }

    @Nested
    class FlushTest {

        @Test
        void shouldWriteFile() throws WritePlayerException {
            when(mockSerializationFactory.createSerializer(any())).thenReturn(mockSerializer);
            when(mockPlayer.uuid()).thenReturn(UUID.randomUUID());
            when(mockSerializer.serialize()).thenReturn("");

            assertEquals(0, testDirectory.listFiles().length);

            fileSystemDB.flush(mockPlayer);

            assertEquals(1, testDirectory.listFiles().length);
        }

        @Test
        void shouldNameFileWithPlayerUUID() throws WritePlayerException {
            final var uuid = UUID.randomUUID();
            when(mockSerializationFactory.createSerializer(any())).thenReturn(mockSerializer);
            when(mockPlayer.uuid()).thenReturn(uuid);
            when(mockSerializer.serialize()).thenReturn("");

            fileSystemDB.flush(mockPlayer);

            final File[] foundFiles = testDirectory.listFiles();
            assertEquals(1, foundFiles.length);
            assertEquals(uuid.toString(), foundFiles[0].getName());
        }

        @Test
        void shouldWriteContentFromSerializer() throws WritePlayerException, IOException {
            when(mockSerializationFactory.createSerializer(any())).thenReturn(mockSerializer);
            when(mockPlayer.uuid()).thenReturn(UUID.randomUUID());
            final String content = "test content from serializer";
            when(mockSerializer.serialize()).thenReturn(content);

            fileSystemDB.flush(mockPlayer);

            final File[] foundFiles = testDirectory.listFiles();
            assertEquals(1, foundFiles.length);
            final String foundContent = Files.readString(foundFiles[0].toPath());
            assertEquals(content, foundContent);
        }
    }

    @Nested
    class FindById {

        @Test
        void shouldThrowIfDirectoryIsAFile() {
            final var notADir = new File(testDirectory, "not_a_dir");
            final var fileSystemDB = new FileSystemPlayerTimeDB(notADir, mockSerializationFactory);
            final Exception exception = assertThrows(RuntimeException.class, () -> fileSystemDB.findById(UUID.randomUUID()));
            assertEquals("Could not find files in: " + notADir.getAbsolutePath(), exception.getMessage());
        }

        @Test
        void shouldBeEmptyIfNoFileIsPresent() throws IOException, ReadPlayerException {
            assertEquals(0, testDirectory.listFiles().length);

            final Optional<Player> foundPlayer = fileSystemDB.findById(UUID.randomUUID());

            assertTrue(foundPlayer.isEmpty());
        }

        @Test
        void shouldBeEmptyIfUuidNotFound() throws IOException, ReadPlayerException {
            new File(testDirectory, "not_a_uuid").createNewFile();

            final Optional<Player> foundPlayer = fileSystemDB.findById(UUID.randomUUID());

            assertTrue(foundPlayer.isEmpty());
        }

        @Test
        void shouldDeserializeFileContent() throws IOException, ReadPlayerException {
            final var playerUUID = UUID.randomUUID();
            final var fakePlayerFile = new File(testDirectory, playerUUID.toString());
            final String fileContent = "test content";
            Files.write(fakePlayerFile.toPath(), fileContent.getBytes());

            when(mockSerializationFactory.createDeserializer(any())).thenReturn(mockPlayer);

            fileSystemDB.findById(playerUUID);

            verify(mockSerializationFactory).createDeserializer(fileContent);
        }
    }

    @Nested
    class FindByName {

        @Test
        void shouldBeEmptyIfNoFilesExist() throws ReadPlayerException {
            final Optional<Player> result = fileSystemDB.findByName("username");
            assertTrue(result.isEmpty());
        }

        @Test
        void shouldFindThroughDeserializePlayerWithOnlyOneFile() throws IOException, ReadPlayerException {
            final var username = "username";
            when(mockPlayer.name()).thenReturn(username);
            final var fakePlayerFile = new File(testDirectory, "fake_player_file");
            Files.write(fakePlayerFile.toPath(), "".getBytes());
            when(mockSerializationFactory.createDeserializer(anyString())).thenReturn(mockPlayer);

            final Optional<Player> result = fileSystemDB.findByName(username);
            assertTrue(result.isPresent());
        }

        @Test
        void shouldNotFindThroughDeserializePlayerWithOnlyOneFile() throws IOException, ReadPlayerException {
            final var username = "username";
            when(mockPlayer.name()).thenReturn(username);
            final var fakePlayerFile = new File(testDirectory, "fake_player_file");
            Files.write(fakePlayerFile.toPath(), "".getBytes());
            when(mockSerializationFactory.createDeserializer(anyString())).thenReturn(mockPlayer);

            final Optional<Player> result = fileSystemDB.findByName("not_a_username");
            assertTrue(result.isEmpty());
        }

        @Test
        void shouldFindThroughDeserializePlayerOnMultipleFiles() throws IOException, ReadPlayerException {
            final var wantedUsername = "username";
            final var notWantedUsername = "not_wanted_username";
            final var rightPlayer = mock(Player.class);
            final var wrongPlayer = mock(Player.class);
            when(rightPlayer.name()).thenReturn(wantedUsername);
            when(wrongPlayer.name()).thenReturn(notWantedUsername);
            final var rightPlayerFileContent = "right";
            final var wrongPlayerFileContent = "wrong";
            final var rightPlayerFile = new File(testDirectory, "right_player_file");
            final var wrongPlayerFile = new File(testDirectory, "wrong_player_file");
            Files.write(rightPlayerFile.toPath(), rightPlayerFileContent.getBytes());
            Files.write(wrongPlayerFile.toPath(), wrongPlayerFileContent.getBytes());
            when(mockSerializationFactory.createDeserializer(rightPlayerFileContent)).thenReturn(rightPlayer);
            when(mockSerializationFactory.createDeserializer(wrongPlayerFileContent)).thenReturn(wrongPlayer);

            final Optional<Player> result = fileSystemDB.findByName(wantedUsername);
            assertTrue(result.isPresent());
        }
    }

    @Nested
    class FindAllTest {

        @Test
        void shouldBeEmptyIfNoFilesExist() throws ReadPlayerException {
            final List<Player> result = fileSystemDB.findAll();
            assertEquals(0, result.size());
        }

        @Test
        void shouldFindOneFile() throws IOException, ReadPlayerException {
            final var fakePlayerFile = new File(testDirectory, "fake_player_file");
            Files.write(fakePlayerFile.toPath(), "test content".getBytes());
            when(mockSerializationFactory.createDeserializer(anyString())).thenReturn(mockPlayer);

            final List<Player> result = fileSystemDB.findAll();
            assertEquals(1, result.size());
        }

        @Test
        void shouldFindTwoFile() throws IOException, ReadPlayerException {
            final var fakePlayerFile = new File(testDirectory, "fake_player_file");
            Files.write(fakePlayerFile.toPath(), "test content".getBytes());
            final var fakePlayerFile2 = new File(testDirectory, "fake_player_file_2");
            Files.write(fakePlayerFile2.toPath(), "test content".getBytes());

            when(mockSerializationFactory.createDeserializer(anyString())).thenReturn(mockPlayer);

            final List<Player> result = fileSystemDB.findAll();
            assertEquals(2, result.size());
        }
    }

}