package de.hhhammer.playtime.ng.serialization.serializer;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import de.hhhammer.playtime.ng.player.PlayTimePlayer;
import de.hhhammer.playtime.ng.player.Player;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;

import java.time.Duration;
import java.time.LocalDateTime;
import java.time.format.DateTimeFormatter;
import java.util.UUID;

import static org.junit.jupiter.api.Assertions.*;

class JsonSerializerTest {
    private final ObjectMapper objectMapper = new ObjectMapper();
    private final Player player = new PlayTimePlayer(UUID.randomUUID(), "username", Duration.ofMinutes(10), LocalDateTime.now());
    private JsonSerializer jsonSerializer;
    private DateTimeFormatter dateTimeFormatter;

    @BeforeEach
    void setUp() {
        dateTimeFormatter = DateTimeFormatter.ISO_DATE_TIME;
        this.jsonSerializer = new JsonSerializer(player, dateTimeFormatter);
    }

    @Test
    void shouldHaveUUID() {
        final var result = jsonSerializer.serialize();
        final String key = "uuid";
        assertTrue(result.contains(key));

        assertEquals(player.uuid().toString(), getValue(key, result));
    }

    @Test
    void shouldHaveUsername() {
        final var result = jsonSerializer.serialize();
        final String key = "name";
        assertTrue(result.contains(key));

        assertEquals(player.name(), getValue(key, result));
    }

    @Test
    void shouldHaveJoinTime() {
        final var result = jsonSerializer.serialize();
        final String key = "joinTime";
        assertTrue(result.contains(key));

        assertEquals(player.joinTime().format(dateTimeFormatter), getValue(key, result));
    }

    @Test
    void shouldReturnPlayerUUID() {
        assertEquals(player.uuid(), jsonSerializer.uuid());
    }

    private String getValue(String key, String json) {
        try {
            var jsonNode = objectMapper.readTree(json);
            return jsonNode.findValue(key).asText();
        } catch (JsonProcessingException e) {
            fail("Finding field failed", e);
        }

        // should never be reached.
        return "";
    }

    @Nested
    class PlaytimeFormattingTest {
        @Test
        void shouldNoPlaytime() {
            final var player = newPlayerWithPlayTime(Duration.ZERO);
            final var jsonSerializer = new JsonSerializer(player, dateTimeFormatter);
            final var result = jsonSerializer.serialize();
            final String key = "playtime";
            assertTrue(result.contains(key));

            assertEquals("00:00:00", getValue(key, result));
        }

        @Test
        void shouldOneSecond() {
            final var player = newPlayerWithPlayTime(Duration.ofSeconds(1));
            final var jsonSerializer = new JsonSerializer(player, dateTimeFormatter);
            final var result = jsonSerializer.serialize();
            final String key = "playtime";
            assertTrue(result.contains(key));

            assertEquals("00:00:01", getValue(key, result));
        }

        @Test
        void shouldTenSeconds() {
            final var player = newPlayerWithPlayTime(Duration.ofSeconds(10));
            final var jsonSerializer = new JsonSerializer(player, dateTimeFormatter);
            final var result = jsonSerializer.serialize();
            final String key = "playtime";
            assertTrue(result.contains(key));

            assertEquals("00:00:10", getValue(key, result));
        }

        @Test
        void shouldOneMinute() {
            final var player = newPlayerWithPlayTime(Duration.ofMinutes(1));
            final var jsonSerializer = new JsonSerializer(player, dateTimeFormatter);
            final var result = jsonSerializer.serialize();
            final String key = "playtime";
            assertTrue(result.contains(key));

            assertEquals("00:01:00", getValue(key, result));
        }

        @Test
        void shouldTenMinutes() {
            final var player = newPlayerWithPlayTime(Duration.ofMinutes(10));
            final var jsonSerializer = new JsonSerializer(player, dateTimeFormatter);
            final var result = jsonSerializer.serialize();
            final String key = "playtime";
            assertTrue(result.contains(key));

            assertEquals("00:10:00", getValue(key, result));
        }

        @Test
        void shouldOneHour() {
            final var player = newPlayerWithPlayTime(Duration.ofHours(1));
            final var jsonSerializer = new JsonSerializer(player, dateTimeFormatter);
            final var result = jsonSerializer.serialize();
            final String key = "playtime";
            assertTrue(result.contains(key));

            assertEquals("01:00:00", getValue(key, result));
        }

        @Test
        void shouldTenHours() {
            final var player = newPlayerWithPlayTime(Duration.ofHours(10));
            final var jsonSerializer = new JsonSerializer(player, dateTimeFormatter);
            final var result = jsonSerializer.serialize();
            final String key = "playtime";
            assertTrue(result.contains(key));

            assertEquals("10:00:00", getValue(key, result));
        }

        @Test
        void shouldHundredHours() {
            final var player = newPlayerWithPlayTime(Duration.ofHours(100));
            final var jsonSerializer = new JsonSerializer(player, dateTimeFormatter);
            final var result = jsonSerializer.serialize();
            final String key = "playtime";
            assertTrue(result.contains(key));

            assertEquals("100:00:00", getValue(key, result));
        }

        private Player newPlayerWithPlayTime(Duration playTime) {
            return new PlayTimePlayer(player.uuid(), player.name(), playTime, player.joinTime());
        }
    }
}
